import time

import triangula.chassis
import triangula.imu
from abc import ABCMeta, abstractmethod
from triangula.input import SixAxis


class TaskManager:
    """
    Manages the task loop
    """

    def __init__(self, arduino, lcd, chassis, joystick):
        self.arduino = arduino
        self.lcd = lcd
        self.chassis = chassis
        self.joystick = joystick
        self.home_task = None

    def _build_context(self, include_bearing):
        bearing = None
        imu_data = None
        if include_bearing:
            imu_data = triangula.imu.read()
            bearing = imu_data[2]
        return TaskContext(arduino=self.arduino,
                           lcd=self.lcd,
                           bearing=bearing,
                           imu_data=imu_data,
                           chassis=self.chassis,
                           joystick=self.joystick,
                           buttons_pressed=self.joystick.get_and_clear_button_press_history())

    def run(self, initial_task):
        """
        Start the task loop. Handles task switching and initialisation as well as any exceptions thrown within tasks.

        :param initial_task:
            An instance of :class:`triangula.task.Task` to use as the first task. Typically this is a menu or startup
            task of some kind.
        """
        active_task = initial_task
        task_initialised = False
        tick = 0

        if self.home_task is None:
            self.home_task = initial_task

        while 1:
            try:
                context = self._build_context(active_task.requires_compass)
                if context.button_pressed(SixAxis.BUTTON_SELECT):
                    active_task = ClearStateTask(self.home_task)
                    task_initialised = False
                    tick = 0
                if task_initialised:
                    new_task = active_task.poll_task(context=context, tick=tick)
                    if new_task is None:
                        tick += 1
                    else:
                        active_task = new_task
                        task_initialised = False
                        tick = 0
                else:
                    active_task.init_task(context=context)
                    task_initialised = True
            except Exception as e:
                active_task = ClearStateTask(ErrorTask(e))
                task_initialised = False


class TaskContext:
    """
    Contains the resources a task might need to perform its function

    :ivar timestamp:
        The time, in seconds since the epoch, that this context was created. In effect this is also the task creation
        time as they're created at the same time.

    """

    def __init__(self, arduino, lcd, bearing, imu_data, chassis, joystick, buttons_pressed):
        """
        Create a new task context

        :param arduino:
            Instance of :class:`triangula.arduino.Arduino` that can be used to manipulate the motors and lights, and
            which can poll the encoders attached to the motors.
        :param lcd:
            Instance of :class:`triangula.lcd.LCD` that can be used to display messages.
        :param bearing:
            If the task has indicated that a bearing is required, this is float value from the compass on the IMU.
        :param imu_data:
            If the task has indicated that a bearing is required, contains the entire IMU data block.
        :param chassis:
            An instance of :class:`triangula.chassis.HoloChassis` defining the motion dynamics for the robot.
        :param joystick:
            An instance of :class:`triangula.input.SixAxis` which can be used to get the joystick axes.
        :param buttons_pressed:
            A bitfield where bits are set to 1 if the corresponding SixAxis button was pressed since the start of the
            previous task poll. Use this in preference to handler registration as it simplifies threading and cleanup
        """
        self.arduino = arduino
        self.lcd = lcd
        self.bearing = bearing
        self.chassis = chassis
        self.joystick = joystick
        self.buttons_pressed = buttons_pressed
        self.timestamp = time.time()
        self.imu_data = imu_data

    def button_pressed(self, button_code):
        """
        Helper method, equivalent to 'self.buttons_pressed & 1 << button_code

        :param button_code:
            A button index from :class:`triangula.input.SixAxis` i.e. SixAxis.BUTTON_SQUARE
        :return:
            0 if the button wasn't pressed at the time the context was created, non-zero otherwise
        """
        return self.buttons_pressed & 1 << button_code


class Task:
    """
    Base class for tasks. Tasks are single-minded activities which are run, one at a time, on Triangula's
    processor. The service script is responsible for polling the active task, providing it with an appropriate
    set of objects and properties such that it can interact with its environment.
    """

    __metaclass__ = ABCMeta

    def __init__(self, task_name='New Task', requires_compass=False):
        """
        Create a new task with the specified name

        :param task_name:
            Name for this task, used in debug mostly. Defaults to 'New Task'
        :param requires_compass:
            Set to True to require that the task is provided with the current compass bearing when polled. This defaults
            to False because I2C transactions are expensive and we don't want to make more of them than we have to.
        """
        self.task_name = task_name
        self.requires_compass = requires_compass

    def __str__(self):
        return 'Task[ task_name={} ]'.format(self.task_name)

    @abstractmethod
    def init_task(self, context):
        """
        Called exactly once, the first time a new task is activated. Use this to set up any properties which weren't
        available during construction.

        :param context:
            An instance of :class:`triangula.task.TaskContext` containing objects and properties which allow the task
            to comprehend and act on its environment.
        """
        return None

    @abstractmethod
    def poll_task(self, context, tick):
        """
        Polled to perform the task's action, you shouldn't hang around too long in this method but there's no explicit
        requirement for timely processing.

        :param context:
            An instance of :class:`triangula.task.TaskContext` containing objects and properties which allow the task
            to comprehend and act on its environment.
        :param int tick:
            A counter, incremented each time poll is called.
        :return:
            Either None, to continue this task, or a subclass of :class:`triangula.task.Task` to switch to that task.
        """
        return None


class ClearStateTask(Task):
    """
    Task which clears the state, turns the lights off and stops the motors, then immediately passes control to another
    task.
    """

    def __init__(self, following_task):
        """
        Create a new clear state task, this will effectively reset the robot's peripherals and pass control to the
        next task. Use this when switching to ensure we're not leaving the wheels running etc.

        :param following_task:
            Another :class:`triangula.task.Task` which is immediately returned from the first poll operation.
        :return:
        """
        super(ClearStateTask, self).__init__(task_name='Clear state task', requires_compass=False)
        self.following_task = following_task

    def init_task(self, context):
        context.arduino.set_motor_power(0, 0, 0)
        context.lcd.set_text(row1='', row2='')
        context.lcd.set_backlight(0, 0, 0)
        context.arduino.set_lights(0, 0, 0)

    def poll_task(self, context, tick):
        return self.following_task


class ErrorTask(Task):
    """
    Task used to display an error message
    """

    def __init__(self, exception):
        """
        Create a new error display task

        :param exception:
            An exception which caused this display to be shown
        """
        super(ErrorTask, self).__init__(task_name='Error', requires_compass=False)
        self.exception = exception
        print exception

    def init_task(self, context):
        context.lcd.set_backlight(red=10, green=0, blue=0)

    def poll_task(self, context, tick):
        context.lcd.set_text(row1='ERROR!', row2=((' ' * 16) + str(self.exception) + (' ' * 16))[
                                                 tick % (len(str(self.exception)) + 16):])
        time.sleep(0.2)



__author__ = 'tom'

from asyncore import file_dispatcher, loop
from threading import Thread

try:
    from evdev import InputDevice, list_devices, ecodes
except ImportError:
    print 'Not importing evdev, expected during sphinx generation on OSX'


class SixAxisResource:
    """
    Resource class which will automatically connect and disconnect to and from a joystick, creating a new SixAxis
    object and passing it to the 'with' clause. Also binds a handler to the START button which resets the axis
    calibration, and to the SELECT button which centres the analogue sticks on the current position.
    """

    def __init__(self, bind_defaults=False, dead_zone=0.05, hot_zone=0.0):
        """
        Resource class, produces a :class:`triangula.input.SixAxis` for use in a 'with' binding.

        :param float dead_zone:
            See SixAxis class documentation
        :param float hot_zone:
            See SixAxis class documentation
        :param bind_defaults:
            Defaults to False, if True will automatically bind two actions to the START and SELECT buttons to
            reset the axis calibration and to set the axis centres respectively.
        """
        self.bind_defaults = bind_defaults
        self.dead_zone = dead_zone
        self.hot_zone = hot_zone

    def __enter__(self):
        self.joystick = SixAxis(dead_zone=self.dead_zone, hot_zone=self.hot_zone)
        self.joystick.connect()
        if self.bind_defaults:
            self.joystick.register_button_handler(self.joystick.reset_axis_calibration, SixAxis.BUTTON_START)
            self.joystick.register_button_handler(self.joystick.set_axis_centres, SixAxis.BUTTON_SELECT)
        return self.joystick

    def __exit__(self, exc_type, exc_value, traceback):
        self.joystick.disconnect()


class SixAxis:
    """
    Class to handle the PS3 SixAxis controller

    This class will process events from the evdev event queue and calculate positions for each of the analogue axes on
    the SixAxis controller (motion sensing is not currently supported). It will also extract
    button press events and call any handler functions bound to those buttons.

    Once the connect() call is made, a thread is created which will actively monitor the device for events, passing them
    to the SixAxis class for processing. There is no need to poll the event queue manually.

    Consuming code can get the current position of any of the sticks from this class through the `axes` instance
    property. This contains a list of :class:`triangula.input.SixAxis.Axis` objects, one for each distinct axis on the
    controller. The list of axes is, in order: left x, left y, right x, right y.
    """

    BUTTON_SELECT = 0  #: The Select button
    BUTTON_LEFT_STICK = 1  #: Left stick click button
    BUTTON_RIGHT_STICK = 2  #: Right stick click button
    BUTTON_START = 3  #: Start button
    BUTTON_D_UP = 4  #: D-pad up
    BUTTON_D_RIGHT = 5  #: D-pad right
    BUTTON_D_DOWN = 6  #: D-pad down
    BUTTON_D_LEFT = 7  #: D-pad left
    BUTTON_L2 = 8  #: L2 lower shoulder trigger
    BUTTON_R2 = 9  #: R2 lower shoulder trigger
    BUTTON_L1 = 10  #: L1 upper shoulder trigger
    BUTTON_R1 = 11  #: R1 upper shoulder trigger
    BUTTON_TRIANGLE = 12  #: Triangle
    BUTTON_CIRCLE = 13  #: Circle
    BUTTON_CROSS = 14  #: Cross
    BUTTON_SQUARE = 15  #: Square
    BUTTON_PS = 16  #: PS button

    def __init__(self, dead_zone=0.05, hot_zone=0.0, connect=False):
        """
        Discover and initialise a PS3 SixAxis controller connected to this computer.

        :param float dead_zone:
            Creates a dead zone centred on the centre position of the axis (which may or may not be zero depending on
            calibration). The axis values range from 0 to 1.0, but will be locked to 0.0 when the measured value less
            centre offset is lower in magnitude than this supplied value. Defaults to 0.05, which makes the PS3 analogue
            sticks easy to centre but still responsive to motion. The deadzone is applies to each axis independently, so
            e.g. moving the stick far right won't affect the deadzone for that sticks Y axis.
        :param float hot_zone:
            Creates a zone of maximum value, any readings from the sensor which are within this value of the max or min
            values will be mapped to 1.0 and -1.0 respectively. This can be useful because, while the PS3 controllers
            sticks have independent axes, they are constrained to move within a circle, so it's impossible to have e.g.
            1.0,1.0 for both x and y axes. Setting this value to non-zero in effect partially squares the circle,
            allowing for controls which require full range control. Setting this value to 1/sqrt(2) will create a square
            zone of variability within the circular range of motion of the controller, with any stick motions outside
            this square mapping to the maximum value for the respective axis. The value is actually scaled by the max
            and min magnitude for upper and lower ranges respectively, so e.g. setting 0.5 will create a hot-zone at
            above half the maximum value and below half the minimum value, and not at +0.5 and -0.5 (unless max and
            min are 1.0 and -1.0 respectively). As with the dead zone, these are applied separately to each axis, so in
            the case where the hot zone is set to 1/sqrt(2), a circular motion of the stick will map to x and y values
            which trace the outline of a square of unit size, allowing for all values to be emitted from the stick.
        :param connect:
            If true, call connect(), otherwise you need to call it elsewhere. Note that connect() may raise IOError if
            it can't find a PS3 controller, so it's best to call it explicitly yourself and handle the error. Defaults
            to False.
        :return: an initialised link to an attached PS3 SixAxis controller.
        """

        self._stop_function = None
        self.axes = [SixAxis.Axis('left_x', dead_zone=dead_zone, hot_zone=hot_zone),
                     SixAxis.Axis('left_y', dead_zone=dead_zone, hot_zone=hot_zone, invert=True),
                     SixAxis.Axis('right_x', dead_zone=dead_zone, hot_zone=hot_zone),
                     SixAxis.Axis('right_y', dead_zone=dead_zone, hot_zone=hot_zone, invert=True)]
        self.button_handlers = []
        self.buttons_pressed = 0
        if connect:
            self.connect()

    def is_connected(self):
        """
        Check whether we have a connection

        :return:
            True if we're connected to a controller, False otherwise.
        """
        if self._stop_function:
            return True
        else:
            return False

    def get_and_clear_button_press_history(self):
        """
        Return the button press bitfield, clearing it as we do.

        :return:
            A bit-field where bits are set to 1 if the corresponding button has been pressed since the last call to
            this method. Test with e.g. 'if button_press_field & SixAxis.BUTTON_CIRCLE:...'
        """
        old_buttons = self.buttons_pressed
        self.buttons_pressed = 0
        return old_buttons

    def connect(self):
        """
        Connect to the first PS3 controller available within /dev/inputX, identifying it by name (this may mean
        that the code doesn't work with non-genuine PS3 controllers, I only have original ones so haven't had
        a chance to test).

        This also creates a new thread to run the asyncore loop, and uses a file dispatcher monitoring the corresponding
        device to handle input events. All events are passed to the handle_event function in the parent, this is then
        responsible for interpreting the events and updating any internal state, calling button handlers etc.

        :return:
            True if a controller was found and connected, False if we already had a connection
        :raises IOError:
            If we didn't already have a controller but couldn't find a new one, this normally means
            there's no controller paired with the Pi
        """
        if self._stop_function:
            return False
        for device in [InputDevice(fn) for fn in list_devices()]:
            if device.name == 'PLAYSTATION(R)3 Controller':
                parent = self

                class InputDeviceDispatcher(file_dispatcher):
                    def __init__(self):
                        self.device = device
                        file_dispatcher.__init__(self, device)

                    def recv(self, ign=None):
                        return self.device.read()

                    def handle_read(self):
                        for event in self.recv():
                            parent.handle_event(event)

                    def handle_error(self):
                        pass

                class AsyncLoop(Thread):
                    def __init__(self, channel):
                        Thread.__init__(self, name='InputDispatchThread')
                        self._set_daemon()
                        self.channel = channel

                    def run(self):
                        loop()

                    def stop(self):
                        self.channel.close()

                loop_thread = AsyncLoop(InputDeviceDispatcher())
                self._stop_function = loop_thread.stop
                loop_thread.start()
                return True
        raise IOError('Unable to find a SixAxis controller')

    def disconnect(self):
        """
        Disconnect from any controllers, shutting down the channel and allowing the monitoring thread to terminate
        if there's nothing else bound into the evdev loop. Doesn't do anything if we're not connected to a controller
        """
        if self._stop_function:
            self._stop_function()
            self._stop_function = None

    def __str__(self):
        """
        Simple string representation of the state of the axes
        """
        return 'x1={}, y1={}, x2={}, y2={}'.format(
            self.axes[0].corrected_value(), self.axes[1].corrected_value(),
            self.axes[2].corrected_value(), self.axes[3].corrected_value())

    def set_axis_centres(self, *args):
        """
        Sets the centre points for each axis to the current value for that axis. This centre value is used when
        computing the value for the axis and is subtracted before applying any scaling.
        """
        for axis in self.axes:
            axis.centre = axis.value

    def reset_axis_calibration(self, *args):
        """
        Resets any previously defined axis calibration to 0.0 for all axes
        """
        for axis in self.axes:
            axis._reset()

    def register_button_handler(self, button_handler, buttons):
        """
        Register a handler function which will be called when a button is pressed

        :param handler: a function which will be called when any of the specified buttons are pressed. The function is
            called with the integer code for the button as the sole argument.
        :param [int] buttons: a list or one or more buttons which should trigger the handler when pressed. Buttons are
            specified as ints, for convenience the PS3 button assignments are mapped to names in SixAxis, i.e.
            SixAxis.BUTTON_CIRCLE. This includes the buttons in each of the analogue sticks. A bare int value is also
            accepted here and will be treated as if a single element list was supplied.
        :return: a no-arg function which can be used to remove this registration
        """
        mask = 0
        if isinstance(buttons, list):
            for button in buttons:
                mask += 1 << button
        else:
            mask += 1 << buttons
        h = {'handler': button_handler,
             'mask': mask}
        self.button_handlers.append(h)

        def remove():
            self.button_handlers.remove(h)

        return remove

    def handle_event(self, event):
        """
        Handle a single evdev event, this updates the internal state of the Axis objects as well as calling any
        registered button handlers.

        :internal:

        :param event:
            The evdev event object to parse
        """
        if event.type == ecodes.EV_ABS:
            value = float(event.value) / 255.0
            if value < 0:
                value = 0
            elif value > 1.0:
                value = 1.0
            if event.code == 0:
                # Left stick, X axis
                self.axes[0]._set(value)
            elif event.code == 1:
                # Left stick, Y axis
                self.axes[1]._set(value)
            elif event.code == 2:
                # Right stick, X axis
                self.axes[2]._set(value)
            elif event.code == 5:
                # Right stick, Y axis (yes, 5...)
                self.axes[3]._set(value)
        elif event.type == ecodes.EV_KEY:
            if event.value == 1:
                if event.code == 288:
                    button = SixAxis.BUTTON_SELECT
                elif event.code == 291:
                    button = SixAxis.BUTTON_START
                elif event.code == 289:
                    button = SixAxis.BUTTON_LEFT_STICK
                elif event.code == 290:
                    button = SixAxis.BUTTON_RIGHT_STICK
                elif event.code == 295:
                    button = SixAxis.BUTTON_D_LEFT
                elif event.code == 292:
                    button = SixAxis.BUTTON_D_UP
                elif event.code == 293:
                    button = SixAxis.BUTTON_D_RIGHT
                elif event.code == 294:
                    button = SixAxis.BUTTON_D_DOWN
                elif event.code == 704:
                    button = SixAxis.BUTTON_PS
                elif event.code == 303:
                    button = SixAxis.BUTTON_SQUARE
                elif event.code == 300:
                    button = SixAxis.BUTTON_TRIANGLE
                elif event.code == 301:
                    button = SixAxis.BUTTON_CIRCLE
                elif event.code == 302:
                    button = SixAxis.BUTTON_CROSS
                elif event.code == 299:
                    button = SixAxis.BUTTON_R1
                elif event.code == 297:
                    button = SixAxis.BUTTON_R2
                elif event.code == 298:
                    button = SixAxis.BUTTON_L1
                elif event.code == 296:
                    button = SixAxis.BUTTON_L2
                else:
                    button = None
                if button is not None:
                    self.buttons_pressed |= 1 << button
                    for button_handler in self.button_handlers:
                        if button_handler['mask'] & (1 << button) != 0:
                            button_handler['handler'](button)

    class Axis():
        """A single analogue axis on the SixAxis controller"""

        def __init__(self, name, invert=False, dead_zone=0.0, hot_zone=0.0):
            self.name = name
            self.centre = 0.5
            self.max = 0.9
            self.min = 0.1
            self.value = 0.5
            self.invert = invert
            self.dead_zone = dead_zone
            self.hot_zone = hot_zone

        def corrected_value(self):
            """
            Get a centre-compensated, scaled, value for the axis, taking any dead-zone into account. The value will
            scale from 0.0 at the edge of the dead-zone to 1.0 (positive) or -1.0 (negative) at the extreme position of
            the controller or the edge of the hot zone, if defined as other than 1.0. The axis will auto-calibrate for
            maximum value, initially it will behave as if the highest possible value from the hardware is 0.9 in each
            direction, and will expand this as higher values are observed. This is scaled by this function and should
            always return 1.0 or -1.0 at the extreme ends of the axis.

            :return: a float value, negative to the left or down and ranging from -1.0 to 1.0
            """

            high_range = self.max - self.centre
            high_start = self.centre + self.dead_zone * high_range
            high_end = self.max - self.hot_zone * high_range

            low_range = self.centre - self.min
            low_start = self.centre - self.dead_zone * low_range
            low_end = self.min + self.hot_zone * low_range

            if self.value > high_start:
                if self.value > high_end:
                    result = 1.0
                else:
                    result = (self.value - high_start) / (high_end - high_start)
            elif self.value < low_start:
                if self.value < low_end:
                    result = -1.0
                else:
                    result = (self.value - low_start) / (low_start - low_end)
            else:
                result = 0

            if not self.invert:
                return result
            else:
                return -result

        def _reset(self):
            """
            Reset calibration (max, min and centre values) for this axis specifically. Not generally needed, you can just
            call the reset method on the SixAxis instance.

            :internal:
            """
            self.centre = 0.5
            self.max = 0.9
            self.min = 0.1

        def _set(self, new_value):
            """
            Set a new value, called from within the SixAxis class when parsing the event queue.

            :param new_value: the raw value from the joystick hardware
            :internal:
            """
            self.value = new_value
            if new_value > self.max:
                self.max = new_value
            elif new_value < self.min:
                self.min = new_value


if __name__ == '__main__':
    from input import SixAxis
    import time

    controller = SixAxis(dead_zone=0.0, hot_zone=0.0)


    def handler(button):
        print 'Button! {}'.format(button)


    controller.register_button_handler(handler, SixAxis.BUTTON_CIRCLE)
    controller.register_button_handler(controller.reset_axis_calibration, SixAxis.BUTTON_START)
    controller.register_button_handler(controller.set_axis_centres, SixAxis.BUTTON_SELECT)

    current_milli_time = lambda: int(round(time.time() * 1000))
    last_time = current_milli_time()
    while 1:
        controller.handle_events()
        now = current_milli_time()
        if now > (last_time + 100):
            last_time = now
            print controller



from math import cos, sin, degrees, radians, pi
from time import time

from euclid import Vector2, Point2
from numpy import array as np_array
from numpy.linalg import solve as np_solve

__author__ = 'tom'


def test():
    chassis = HoloChassis(wheels=[
        HoloChassis.OmniWheel(position=Point2(1, 0), angle=0, radius=60),
        HoloChassis.OmniWheel(position=Point2(-1, 0), angle=0, radius=60)]
    )
    print chassis.get_wheel_speeds(Motion(translation=Vector2(0, 0), rotation=0.5))
    print chassis.get_wheel_speeds(Motion(translation=Vector2(0, 0), rotation=0.5), origin=Point2(1, 0))


def rotate_point(point, angle, origin=None):
    """
    Rotate a Point2 around another Point2

    :param euclid.Point2 point:
        The point to rotate
    :param float angle:
        Angle in radians, clockwise rotation
    :param euclid.Point2 origin:
        Origin of the rotation, defaults to (0,0) if not specified
    :return:
        A new :class:`euclid.Point2` containing the rotated input point
    """
    if origin is None:
        origin = Point2(0, 0)
    s = sin(-angle)
    c = cos(-angle)
    return Point2(c * (point.x - origin.x) - s * (point.y - origin.y) + origin.x,
                  s * (point.x - origin.x) + c * (point.y - origin.y) + origin.y)


def rotate_vector(vector, angle, origin=None):
    """
    Rotate a :class:`euclid.Vector2` around a :class:`euclid.Point2`

    :param euclid.Vector2 vector:
        The vector to rotate
    :param float angle:
        Angle in radians, clockwise rotation
    :param euclid.Point2 origin:
        Origin of the rotation, defaults to (0,0) if not specified
    :return:
        A new :class:`euclid.Point2` containing the rotated input point
    """
    if origin is None:
        origin = Point2(0, 0)
    s = sin(-angle)
    c = cos(-angle)
    return Vector2(c * (vector.x - origin.x) - s * (vector.y - origin.y) + origin.x,
                   s * (vector.x - origin.x) + c * (vector.y - origin.y) + origin.y)


def smallest_difference(a, b, max_value=2 * pi):
    """
    Given two floats, a and b, and a maximum possible value for both a and b, calculate the smallest delta from a to b.
    For example, if a=1.0, b=2.5 and max_value=2.6, this should return -1.1, as subtracting 1.1 from a would result in
    -0.1, which will then be transformed to 2.5 after taking its modulus with 2.6. If max_value was 10, it would return
    +1.5, as this is the lower magnitude delta needed to go from 1.0 to 2.5. This function is used when calculating the
    shortest delta between two pose orientations, for this reason the max_value defaults to 2*pi for use when working
    in radians.

    If either a or b are less than zero or greater than the maximum value they will be treated as a % max_value or b %
    max_value respectively for the purposes of this calculation.

    :param float a:
        First value (see above)
    :param b:
        Second value (see above)
    :param max_value:
        Modulus, defaults to 2*pi if not specified
    :return:
        A value d such that (a + d) % max_value == b, and abs(d) is minimal (as there would be an infinite number of
        possible d that satisfy this relationship).
    """
    mod_a = a % max_value
    mod_b = b % max_value
    if abs(mod_a - mod_b) <= max_value / 2:
        return mod_b - mod_a
    elif mod_a >= mod_b:
        return max_value - (mod_a + mod_b)
    else:
        return mod_b + mod_a - max_value


def get_regular_triangular_chassis(wheel_distance, wheel_radius, max_rotations_per_second):
    """
    Build a HoloChassis object with three wheels, each identical in size and maximum speed. Each wheel is positioned
    at the corner of a regular triangle, and with direction perpendicular to the normal vector at that corner.

    :param wheel_distance:
        Distance in millimetres between the contact points of each pair of wheels (i.e. the length of each edge of the
        regular triangle)
    :param wheel_radius:
        Wheel radius in millimetres
    :param max_rotations_per_second:
        Maximum wheel speed in revolutions per second
    :return:
        An appropriately configured HoloChassis
    """
    point = Point2(0, cos(radians(30)) * wheel_distance / 2.0)
    vector = Vector2(-2 * pi * wheel_radius, 0)

    # Pink
    wheel_a = HoloChassis.OmniWheel(
        position=point,
        vector=vector,
        max_speed=max_rotations_per_second)
    # Yellow
    wheel_b = HoloChassis.OmniWheel(
        position=rotate_point(point, pi * 2 / 3),
        vector=rotate_vector(vector, pi * 2 / 3),
        max_speed=max_rotations_per_second)
    # Green
    wheel_c = HoloChassis.OmniWheel(
        position=rotate_point(point, pi * 4 / 3),
        vector=rotate_vector(vector, pi * 4 / 3),
        max_speed=max_rotations_per_second)

    return HoloChassis(wheels=[wheel_a, wheel_b, wheel_c])


class WheelSpeeds:
    """
    A simple container to hold desired wheel speeds, and to indicate whether any speeds were scaled back due to
    impossibly high values.
    """

    def __init__(self, speeds, scaling):
        """
        Create a new wheel speeds container

        :param speeds:
            A sequence of float values, one per wheel, in revolutions per second
        :param float scaling:
            If a requested translation or rotation was too fast for the chassis to perform, it will return an instance
            of this class with the scaling set to a value greater than 1.0. This indicates that it was unable to
            provide the requested trajectory but has instead provided the highest magnitude one possible. This parameter
            then contains the proportion of the requested trajectory that was possible to provide. For example, if
            the motion requested was a translation of 10mm/s in the X axis and a rotation of 10 radians per second, but
            on calculation this resulted in excessive wheel speeds which weren't possible, it might be scaled back to
            6mm/s on X and 6 radians per second - the motion is proportionately the same just slower, and in this case
            the scaling value would be 0.6.
        """
        self.speeds = speeds
        self.scaling = scaling

    def __str__(self):
        return 'WheelSpeeds[ speeds={}, scaling={} ]'.format(self.speeds, self.scaling)


class Motion:
    """
    A container to hold the translation and rotation vector representing the robot's motion. This is always expressed
    in the robot's coordinate frame, so a translation component of 0,1 always means the robot is heading forwards,
    irrespective of the current orientation of the robot (i.e. if the robot was turned 90 degrees in world space this
    0,1 motion would be a movement along the X axis in world space, but the Y axis in robot space). The rotation
    component of the motion is expressed in radians per second, positive values corresponding to clockwise rotation
    when viewed from the direction relative to the plane such that X is positive to the right and Y positive upwards.
    """

    def __init__(self, translation=None, rotation=0):
        """
        Constructor

        :param euclid.Vector2 translation:
            Vector2 representing the translation component in robot coordinate space of the motion. Defaults to
            Vector2(0,0)
        :param float rotation:
            Rotation in radians per second. Defaults to 0.
        """
        if translation is not None:
            self.translation = translation
        else:
            self.translation = Vector2(0, 0)
        self.rotation = rotation

    def __str__(self):
        return 'Motion[ x={}, y={}, theta={} (deg={}) ]'.format(self.translation.x, self.translation.y, self.rotation,
                                                                degrees(self.rotation))


class DeadReckoning:
    """
    Encapsulates the logic required to track the robot's position in world space using wheel encoders and chassis
    kinematics. To update the state of this object you need to call the update_from_counts function - this will
    compute the difference in counts for each wheel, and from this derive the rotational speed for each wheel since
    the last measurement. The :class:`triangula.chassis.HoloChassis` is then used to convert these speeds into an arc,
    with the assumption that wheel speeds were constant during the time interval. This arc is used to update the
    :class:`triangula.chassis.Pose` representing the current best estimate of the robot's position.

    Because this is in effect integrating over sensor readings, any errors, particularly in the chassis geometry or
    dimensions, or in the number of counts per revolution (for example if the gearing isn't quite what you think it is
    or there's enough slop in the gearbox that readings can drift) will accumulate over time. To mitigate this, if you
    have precise instantaneous information such as a compass reading every few seconds, these readings can be used to
    explicitly set the position, orientation, or both of the :class:`triangula.chassis.Pose` tracked by this class.

    As there's an implicit assumption that wheel speeds are constant between encoder readings, this class will yield
    more accurate results when updated frequently. The exact optimal update frequency will depend on the encoder
    resolutions, chassis geometry etc. Some manual tuning may be required.
    """

    def __init__(self, chassis, counts_per_revolution=64 * 19, max_count_value=1 << 15):
        """
        Constructor

        :param triangula.chassis.HoloChassis chassis:
            The :class:`triangula.chassis.HoloChassis` to be used to define kinematics for this DeadReckoning
        :param float counts_per_revolution:
            The number of counts registered by the wheel encoders per revolution of the wheel. Defaults to 64*19 to
            be the 64 count encoder fitted to a 19:1 reduction gearbox.
        :param int max_count_value:
            The largest value read from the encoders, this is used to determine when we've wrapped around the zero
            point, defaults to 1<<16 to reflect that count values are held in the microcontroller module as a uint16_t
        """
        self.chassis = chassis
        self.counts_per_revolution = counts_per_revolution
        self.max_count_value = max_count_value
        self.last_encoder_values = None
        self.last_reading_time = None
        self.pose = None

    def reset(self):
        """
        Clear the state of this :class:`triangula.chassis.DeadReckoning`
        """
        self.last_encoder_values = None
        self.last_reading_time = None
        self.pose = None

    def set_position(self, position):
        """
        Explicitly set the position of the robot in world coordinates. Overrides the current value tracked by this
        instance. Use this when you have better information and want to update the state accordingly.

        :param euclid.Point2 position:
            The new position to set, as a :class:`euclid.Point2`, coordinates are in mm
        """
        self.pose.position = position
        return self.pose

    def set_orientation(self, orientation):
        """
        Explicitly set the orientation of the robot in world coordinates. Use this to explicitly update the orientation,
        for example when you have a sufficiently accurate compass fix that it can be used to eliminate any accumulated
        errors built up by the dead reckoning algorithm.

        :param float orientation:
            The new orientation to set, in radians from the positive Y axis, clockwise rotations being positive. This
            value will be normalised to the range 0-2PI
        :return:
            The current (updated) value of the :class:`triangula.chassis.Pose`
        """
        self.pose.orientation = orientation % (2 * pi)
        return self.pose

    def update_from_counts(self, counts):
        """
        Update the pose from a new set of encoder values

        :param counts:
            A list of encoder counts, one per wheel
        :return:
            The updated :class:`triangula.chassis.Pose` object (this is also modified in the internal state of the
            DeadReckoning)
        """
        reading_time = time()
        if self.last_encoder_values is None:
            self.last_encoder_values = counts
            self.last_reading_time = reading_time
            self.pose = Pose(Point2(0, 0), 0)
        else:
            time_delta = reading_time - self.last_reading_time
            wheel_speeds = [smallest_difference(current_reading, last_reading, self.max_count_value) / (
                self.counts_per_revolution * time_delta) for last_reading, current_reading
                            in zip(counts, self.last_encoder_values)]
            motion = self.chassis.calculate_motion(speeds=wheel_speeds)
            self.pose = self.pose.calculate_pose_change(motion, time_delta)
            self.last_encoder_values = counts
            self.last_reading_time = reading_time
        return self.pose


class Pose:
    """
    A container to hold the position as a Point2 along with orientation in radians, where 0 corresponds to the positive
    Y axis (0,1). Orientation is expressed in radians, with positive values indicating a rotation from the positive Y
    axis in the clockwise direction, i.e. a rotation of 0 is North, pi/2 East, pi South and 3pi/2 West.
    """

    def __init__(self, position=None, orientation=0):
        """
        Constructor

        :param euclid.Point2 position:
            A Point2 containing the position of the centre of the robot. Defaults to Point2(0,0)
        :param float orientation:
            Orientation in radians, 0 being the positive Y axis, positive values correspond to clockwise rotations, i.e.
            pi/4 is East. This value will be normalised to be between 0 and 2 * pi. Defaults to 0
        """
        if position is not None:
            self.position = position
        else:
            self.position = Point2(0, 0)
        self.orientation = orientation % (2 * pi)

    def distance_to_pose(self, to_pose):
        """
        Return the distance to the other pose position

        :param triangula.chassis.Pose to_pose:
            The target pose
        """
        return abs(self.position - to_pose.position)

    def is_close_to(self, to_pose, max_distance=0.001, max_orientation_difference=radians(1)):
        """
        Check whether we're close to the specified pose, defining closeness as both distance on the plane and difference
        in orientation.

        :param to_pose:
            The target pose
        :param max_distance:
            Maximum distance within which we'll count as being close, defaults to 0.001
        :param max_orientation_difference:
            Maximum number of radians we can be off the target pose's orientation to count as close, defaults to 1
            degree (calculated with ``radians(1)``)
        :return:
            True if this pose is regarded as close to the other, False otherwise
        """
        if self.distance_to_pose(to_pose) > max_distance:
            return False
        elif smallest_difference(self.orientation, to_pose.orientation) > max_orientation_difference:
            return False
        else:
            return True

    def translate(self, vector):
        """
        Create a new pose, with the same orientation as this one and the specified translation applied to its position.

        :param euclid.Vector2 vector:
            Vector by which the position of this pose should be translated when creating the new Pose
        :return:
            Returns the new Pose
        """
        return Pose(position=self.position + vector, orientation=self.orientation)

    def pose_to_pose_vector(self, to_pose):
        """
        Calculates the Vector2, in robot coordinate space (remember that Pose objects use world coordinates!) that
        represents the translation required to move from this Pose to the specified target Pose.

        :param triangula.chassis.Pose to_pose:
            A target :class:`triangula.chassis.Pose`, the resultant vector in robot space will translate the robot to
            the position contained in this pose. Note that this does not take any account of the orientation component
            of the to_pose, only the starting one.
        :return:
            A :class:`euclid.Vector2` containing the translation part, in robot space, of the motion required to move
            from this Pose to the target.
        """
        return rotate_vector(
            vector=Vector2(to_pose.position.x - self.position.x, to_pose.position.y - self.position.y),
            angle=-self.orientation)

    def calculate_pose_change(self, motion, time_delta):
        """
        Given this as the starting Pose, a Motion and a time in seconds, calculate the resultant Pose at the end of the
        time interval.

        This makes use of the fact that if you travel in a consistent direction while turning at a constant rate you
        will describe an arc. By calculating the centre point of this arc we can simply rotate the starting pose around
        this centre point. This is considerably simpler than integrating over the motion 3-vector. A special case is
        used to avoid division by zero errors when there is no rotation component to the motion.

        :param triangula.chassis.Motion motion:
            The motion of the robot, assumed to be constant for the duration of the time interval. The motion is
            expressed in the robot's coordinate frame, so a translation of (0,1) is always a forward motion,
            irrespective of the current orientation.
        :param float time_delta:
            The time in seconds during which the specified motion should be applied.
        :return:
            A :class:`triangula.chassis.Pose` which represents resultant pose after applying the supplied motion for the
            given time.
        """

        # Total delta in orientation angle over the time interval
        orientation_delta = motion.rotation * time_delta
        # Scaled translation vector rotated into world coordinate space (motion uses robot space)
        translation_vector_world = rotate_vector(motion.translation, self.orientation) * time_delta
        ':type : euclid.Vector2'

        if orientation_delta == 0:
            # No orientation, trivially add the rotated, scaled, translation vector to the current pose
            return self.translate(translation_vector_world)
        else:
            centre_of_rotation = self.position + translation_vector_world.cross() / orientation_delta
            ':type : euclid.Point2'
            final_position = rotate_point(self.position, angle=orientation_delta, origin=centre_of_rotation)
            return Pose(position=final_position, orientation=self.orientation + orientation_delta)

    def __str__(self):
        return 'Pose[x={}, y={}, orientation={} (deg={})]'.format(self.position.x, self.position.y, self.orientation,
                                                                  degrees(self.orientation))


class HoloChassis:
    """
    An assembly of wheels at various positions and angles, which can be driven independently to create a holonomic drive
    system. A holonomic system is one where number of degrees of freedom in the system is equal to the number of
    directly controllable degrees of freedom, so for a chassis intended to move in two dimensions the degrees of freedom
    are two axes of translation and one of rotation. For a full holonomic system we therefore need at least three wheels
    defined.
    """

    def __init__(self, wheels):
        """
        Create a new chassis, specifying a set of wheels.
        
        :param wheels:
            A sequence of :class:`triangula.chassis.HoloChassis.OmniWheel` objects defining the wheels for this chassis.
        """
        self.wheels = wheels
        self._matrix_coefficients = np_array([[wheel.co_x, wheel.co_y, wheel.co_theta] for wheel in self.wheels])

    def calculate_motion(self, speeds):
        """
        Invert the motion to speed calculation to obtain the actual linear and angular velocity of the chassis given
        a vector of wheel speeds. See http://docs.scipy.org/doc/numpy-1.10.1/reference/generated/numpy.linalg.solve.html

        :param speeds:
            An array of wheel speeds, expressed as floats with units of radians per second, positive being towards
            the wheel vector.
        :return:
            A :class:`triangula.chassis.Motion` object containing the calculated translation and rotation in the robot's
            coordinate space.
        """
        motion_array = np_solve(self._matrix_coefficients, np_array(speeds))
        return Motion(Vector2(x=float(motion_array[0]),
                              y=float(motion_array[1])),
                      rotation=float(motion_array[2]))

    def get_max_translation_speed(self):
        """
        Calculate the maximum translation speed, assuming all directions are equivalent and that there is no rotation
        component to the motion.

        :return:
            Maximum speed in millimetres per second as a float
        """
        unrealistic_speed = 10000.0
        scaling = self.get_wheel_speeds(Motion(translation=Vector2(0, unrealistic_speed), rotation=0)).scaling
        return unrealistic_speed * scaling

    def get_max_rotation_speed(self):
        """
        Calculate the maximum rotation speed around the origin in radians per second, assuming no translation motion
        at the same time.

        :return:
            Maximum radians per second as a float
        """
        unrealistic_speed = 2 * pi * 100
        scaling = self.get_wheel_speeds(Motion(translation=Vector2(0, 0), rotation=unrealistic_speed)).scaling
        return unrealistic_speed * scaling

    def get_wheel_speeds(self, motion, origin=Point2(x=0, y=0)):
        """
        Calculate speeds to drive each wheel in the chassis at to attain the specified rotation / translation 3-vector.

        :param triangula.chassis.Motion motion:
            Desired motion of the robot chassis
        :param euclid.Point2 origin:
            Optional, can define the centre of rotation to be something other than 0,0. Units are in millimetres.
            Defaults to rotating around x=0, y=0.
        :return:
            A :class:`triangula.chassis.WheelSpeeds` containing both the target wheel speeds and the scaling, if any,
            which was required to bring those speeds into the allowed range for all wheels. This prevents unexpected
            motion in cases where only a single wheel is being asked to turn too fast, in such cases all wheel speeds
            will be scaled back such that the highest is within the bounds allowed for that particular wheel. This
            can accommodate wheels with different top speeds.
        """

        def velocity_at(point):
            """
            Compute the velocity as a Vector2 at the specified point given the enclosing translation and rotation values

            Method: Normalise the vector from the origin to the point, then take the cross of itself to produce a unit
            vector with direction that of a rotation around the origin. Scale this by the distance from the origin and
            by the rotation in radians per second, then simply add the translation vector.

            :param euclid.Point2 point:
                Point at which to calculate velocity
            :return:
                A :class:`euclid.Vector2` representing the velocity at the specified point in mm/s
            """
            d = point - origin
            return d.cross() * motion.rotation + motion.translation

        wheel_speeds = list(wheel.speed(velocity_at(wheel.position)) for wheel in self.wheels)
        scale = 1.0
        for speed, wheel in zip(wheel_speeds, self.wheels):
            if wheel.max_speed is not None and abs(speed) > wheel.max_speed:
                wheel_scale = wheel.max_speed / abs(speed)
                scale = min(scale, wheel_scale)
        return WheelSpeeds(speeds=list(speed * scale for speed in wheel_speeds), scaling=scale)

    class OmniWheel:
        """
        Defines a single omni-wheel within a chassis assembly. Omni-wheels are wheels formed from rollers, where the
        motion of the roller is perpendicular to the motion of the primary wheel. This is distinct from a mechanum wheel
        where the rollers are at an angle (normally around 40-30 degrees) to the primary wheel. Omni-wheels must be
        positioned on the chassis with non-parallel unit vectors, mechanum wheels can in some cases be positioned with
        all unit vectors parallel.

        A wheel has a location relative to the chassis centre and a vector describing the direction of motion of the
        wheel when driven with a positive angular velocity. The location is specified in millimetres, and the magnitude
        of the wheel vector should be equal to the number of millimetres travelled in a single revolution. This allows
        for different sized wheels to be handled within the same chassis.
        """

        def __init__(self, position, max_speed=0, angle=None, radius=None, vector=None):
            """
            Create a new omni-wheel object, specifying the position and either a direction vector directly or the angle
            in degrees clockwise from the position Y axis along with the radius of the wheel.

            :param euclid.Point2 position:
                The wheel's contact point with the surface, specified relative to the centre of the
                chassis. Units are millimetres.
            :param float max_speed:
                The maximum number of revolutions per second allowed for this wheel. When calculating the wheel speeds
                required for a given trajectory this value is used to scale back all motion if any wheel would have to
                move at an impossible speed. If not specified this defaults to None, indicating that no speed limit
                should be placed on this wheel.
            :param angle:
                The angle, specified in radians from the positive Y axis where positive values are clockwise from this
                axis when viewed from above, of the direction of travel of the wheel when driven with a positive speed.
                If this value is specified then radius must also be specified and dx,dy left as None.
            :param radius:
                The radius in millimetres of the wheel, measuring from the centre to the contact point with the surface,
                this may be hard to determine for some wheels based on their geometry, particularly for wheels with
                cylindrical rollers, as the radius will vary. For these cases it may be worth directly measuring the
                circumference of the entire assembly and calculating radius rather than measuring directly. This is used
                to determine the magnitude of the direction vector. If this is not None then the angle must also be
                specified, and dx,dy left as None.
            :param euclid.Vector2 vector:
                2 dimensional vector defining the translation of the wheel's contact point after a full
                revolution of the wheel.
            """
            self.position = position
            self.max_speed = max_speed
            if angle is None and radius is None and vector is not None:
                #  Specify wheel based on direct vector """
                self.vector = vector
            elif angle is not None and radius is not None and vector is None:
                # Specify based on angle from positive Y axis and radius """
                circumference = 2 * pi * radius
                self.vector = Vector2(sin(angle) * circumference, cos(angle) * circumference)
            else:
                raise ValueError('Must specify exactly one of angle and radius or translation vector')
            self.vector_magnitude_squared = self.vector.magnitude_squared()
            self.co_x = self.vector.x / self.vector_magnitude_squared
            self.co_y = self.vector.y / self.vector_magnitude_squared
            self.co_theta = (self.vector.x * self.position.y -
                             self.vector.y * self.position.x) / self.vector_magnitude_squared

        def speed(self, velocity):
            """
            Given a velocity at a wheel contact point, calculate the speed in revolutions per second at which the wheel
            should be driven.

            Method: we want to find the projection of the velocity onto the vector representing the drive of this wheel.
            We store the vector representing a single revolution of travel as self.vector, so the projection onto this
            would be velocity.dot(self.vector / abs(self.vector)). However, we want revolutions per second, so we must
            then divide again by abs(self.vector), leading to
            velocity.dot(self.vector / abs(self.vector))/abs(self.vector). Because the definition of the dot product is
            the sum of x1*x2, y1*y2, ... any scalar applied to each x, y ... of a single vector can be moved outside
            the dot product, so we can simplify as velocity.dot(self.vector) / abs(self.vector)^2. As the magnitude of
            the vector is taken by sqrt(x^2+y^2) we can simply express this as (x^2+y^2), held in the convenient
            function magnitude_squared(). So our final simplified form is
            velocity.dot(self.vector) / self.vector.magnitude_squared(). For efficiency, and because self.vector doesn't
            change, we can pre-compute this.

            :param euclid.Vector2 velocity:
                The velocity at the wheel's contact point with the surface, expressed in mm/s
            :return:
                Target wheel speed in rotations per second to hit the desired vector at the contact point.
            """
            return velocity.dot(self.vector) / self.vector_magnitude_squared



from time import time as time_now

from triangula.chassis import Motion


class MotionLimit:
    """
    Utility class to limit rate of motion changes, similar to the :class:`triangula.dynamics.RateLimit` but specialised
    to limit rate of change of :class:`triangula.chassis.Motion` instances. It computes the necessary linear and angular
    acceleration required to transition from the previously registered Motion to the new one, and scales between the two
    Motions such that the limits aren't exceeded. Linear interpolation is used to map between the previous and new
    motions. This works around the drawback with the :class:`triangula.dynamics.RateLimit` when applied to wheel speeds
    where the resultant motion will incorporate a degree of rotation even when no rotations are included in either
    motion.
    """

    def __init__(self, linear_acceleration_limit, angular_acceleration_limit):
        """
        Create a new instance configured with the specified limits.

        :param float linear_acceleration_limit:
            Maximum allowed linear acceleration in mm per second per second
        :param angular_acceleration_limit:
            Maximum allowed radial acceleration in radians per second per second
        """
        self.linear_acceleration_limit = linear_acceleration_limit
        self.angular_acceleration_limit = angular_acceleration_limit
        self.last_motion = None
        self.last_motion_time = None

    def limit_and_return(self, motion):
        """
        Apply limits to the requested motion based on the current state of the MotionLimit, returning the closest Motion
        which complies with the specified limits.

        :param triangula.chassis.Motion motion:
            The requested :class:`triangula.chassis.Motion`,
        :return:
            The modified motion, or the supplied one if it complied with the limits
        """
        now = time_now()
        if self.last_motion is None:
            self.last_motion = motion
            self.last_motion_time = now
            return motion
        # Calculate the requested linear acceleration magnitude in mm/s/s to achieve the desired motion change
        time_delta = now - self.last_motion_time
        motion_delta = abs(motion.translation - self.last_motion.translation)
        linear_acceleration = motion_delta / time_delta
        angular_acceleration = abs(motion.rotation - self.last_motion.rotation) / time_delta
        scaling = 1.0
        if linear_acceleration > self.linear_acceleration_limit:
            scaling = self.linear_acceleration_limit / linear_acceleration
        if angular_acceleration > self.angular_acceleration_limit:
            scaling = min(scaling, self.angular_acceleration_limit / angular_acceleration)
        scaled_translation = motion.translation * scaling + self.last_motion.translation * (1.0 - scaling)
        ':type : euclid.Vector2'
        scaled_rotation = motion.rotation * scaling + self.last_motion.rotation * (1.0 - scaling)
        scaled_motion = Motion(rotation=scaled_rotation, translation=scaled_translation)
        self.last_motion = scaled_motion
        self.last_motion_time = now
        return scaled_motion


class RateLimit:
    """
    Utility class to provide time-based rate limiting.
    """

    def __init__(self, limit_function=None):
        """
        Create a new rate limit object, this can be used to enforce maximum rates of change in a set of values, these
        are generally expressed in rate per second, but could use any arbitrary function. For a default application the
        provided function which can handle rate per second should be sufficient, but the ability to provide a custom
        function allows for e.g. permitting larger rates when decreasing. This can be used in conjunction with the motor
        power functions to provide a degree of traction control (we can't actually detect slipping in wheels, we just
        don't have sufficient information), by limiting the maximum requested power rate change over time.

        :param limit_function:
            A function of old_value * old_time * new_value * new_time which will return a potentially limited new value
            given a requested value, historical value and timepoints for both.
        """
        self.previous_values = None
        self.previous_time = None
        self.limit_function = limit_function

    def limit_and_return(self, values):
        """
        Take a list of values, update the internal state of the RateLimit and return a modified list of values which are
        restricted by the configured limit function.

        :param float[] values:
            Values to attempt to apply
        :return:
            New values to apply, modified by the configured limit function
        """
        now = time_now()
        if self.previous_time is None:
            self.previous_time = now
            self.previous_values = values
            return values
        updated_values = [self.limit_function(previous_value, self.previous_time, value, now) for
                          (previous_value, value) in
                          zip(self.previous_values, values)]
        self.previous_values = updated_values
        self.previous_time = now
        return updated_values

    @staticmethod
    def fixed_rate_limit_function(rate_per_second):
        """
        Create and return a new limit function which will lock the maximum delta applied to the specified rate per
        second.

        :param float rate_per_second:
            Largest allowed delta per second
        :return:
            A function which can be used in the :class:`triangula.dynamics.RateLimit` constructor and which will enforce
            a fixed maximum absolute rate of change across successive values such that no value in the supplied vector
            will vary at a higher rate than provided.
        """

        def limit_function(previous_value, previous_time, new_value, new_time):
            time_delta = new_time - previous_time
            value_delta = abs(previous_value - new_value)
            requested_rate = value_delta / time_delta
            if requested_rate <= rate_per_second:
                return new_value
            else:
                return previous_value + ((new_value - previous_value) * rate_per_second / requested_rate)

        return limit_function



__author__ = 'tom'

from time import sleep

try:
    import smbus
except ImportError:
    print 'Not importing smbus, expected during sphinx generation on OSX'

ARDUINO_ADDRESS = 0x70

I2C_RETRIES = 30
I2C_DELAY = 0.01


def float_to_byte(f):
    """
    Map a float from -1.0 to 1.0 onto the range 0-255 and return as a byte-compatible number.
    Used when sending bytes for e.g. the speed controllers.

    :param f:
        A float, between -1.0 and 1.0. Values outside this range will not behave sensibly.
    :return:
        An integer where 128 is equivalent to 0, 0 to -1.0 and 255 to +1.0
    """
    i = int((f + 1.0) * 128.0)
    return check_byte(i)


def check_byte(b):
    """
    Clamp the supplied value to an integer between 0 and 255 inclusive

    :param b:
        A number
    :return:
        Integer representation of the number, limited to be between 0 and 255
    """
    i = int(b)
    if i < 0:
        i = 0
    elif i > 255:
        i = 255
    return i


def compute_checksum(register, data):
    """
    Calculate a checksum for the specified data and register, simply by XORing each byte in turn with the current value
    of the checksum. The value is initialised to be the register, then each byte of the data array is used in turn. This
    is intended to inter-operate with the code on the Arduino, it's not a particularly generic checksum routine and
    won't work with arbitrary I2C devices.

    :param int register:
        The register to which we're sending the data. This is used because the first part of the data transaction
        is actually the register, we want to calculate the checksum including the register part.
    :param int[] data:
        The data array to use
    :return:
        A checksum, this should be appended onto the data sent to the Arduino
    """
    xor = register
    for data_byte in data:
        xor ^= data_byte
    return xor


class Arduino:
    """
    Handles communication over I2C with the Arduino, exposing methods which can be used to
    read e.g. wheel encoder positions, and to set wheel speeds, light colours etc.

    Client code based on documentation at
    http://www.raspberry-projects.com/pi/programming-in-python/i2c-programming-in-python/using-the-i2c-interface-2

    The logic for programming the Arduino to respond to these messages is implemented
    elsewhere, and based on the code at http://dsscircuits.com/articles/arduino-i2c-slave-guide.
    """

    DEVICE_MOTORS_SET = 0x20
    'Register used to set motor power'
    DEVICE_LIGHTS_SET = 0x21
    'Register used to set the lights to a constant colour'
    DEVICE_ENCODERS_READ = 0x22
    'Register used to copy the current encoder values into the read buffer on the Arduino'

    def __init__(self,
                 arduino_address=ARDUINO_ADDRESS,
                 i2c_delay=I2C_DELAY,
                 max_retries=I2C_RETRIES,
                 bus_id=1):
        """
        Create a new client, uses the I2C bus to communicate with the arduino.

        :param arduino_address:
            7-bit address of the arduino on the I2C bus
        :param i2c_delay:
            Delay in seconds, used when we either retry transactions or inbetween immediately
            consecutive read and write operations. This attempts to mitigate problems we've had
            where transactions have failed, probably because the arduino is still handling the
            previous read / write mode and gets confused.
        :param max_retries:
            The maximum number of retries to attempt when an IOError occurs during a transaction.
        :param bus_id:
            The bus ID used when initialising the smbus library. For the Pi2 this is 1, others may
            be on bus 0.
        """
        self._bus = smbus.SMBus(bus_id)
        self._i2c_delay = i2c_delay
        self._max_retries = max_retries
        self._address = arduino_address

    def _send(self, register, data):
        retries_left = self._max_retries
        while retries_left > 0:
            try:
                data_with_checksum = []
                data_with_checksum.extend(data)
                data_with_checksum.append(compute_checksum(register, data))
                self._bus.write_i2c_block_data(self._address, register, data_with_checksum)
                return
            except IOError:
                sleep(self._i2c_delay)
                retries_left -= 1
                pass
        raise IOError("Retries exceeded sending data to arduino.")

    def _read(self, register, bytes_to_read):
        """
        Work around issues we seem to have with bulk transfers by transferring a byte at
        a time after manually triggering the population of the read buffer on the arduino.

        :param register:
            The register to write a 0 to in order to start the read transaction. The slave
            device should respond to this by populating the export array, this is then read
            a byte at a time.
        :param bytes_to_read:
            The number of times the read_byte method needs to be called
        :return:
            A byte array, length equal to bytes_to_read, containing data read from the arduino.
        """
        retries_left = self._max_retries
        while retries_left > 0:
            try:
                # Prod the appropriate control register
                self._send(register, [0])
                # Delay for an arbitrary amount of time
                sleep(self._i2c_delay * 10)
                # Call read_byte repeatedly to assemble our output data
                data = [self._bus.read_byte(self._address) for _ in xrange(bytes_to_read)]
                return data
            except IOError:
                sleep(self._i2c_delay)
                retries_left -= 1
        raise IOError("Retries exceeded when fetching data from arduino.")

    def set_motor_power(self, a, b, c):
        """
        Set motor power, writing values directly to the Syren controllers. Power values range from -1.0 to 1.0, where
        positive values correspond to increasing encoder counts and clockwise rotation when viewed from the outside
        of the robot looking inwards.

        :param float a:
            Wheel a power, -1.0 to 1.0. Pink pylon.
        :param float b:
            Wheel a power, -1.0 to 1.0. Yellow pylon.
        :param float c:
            Wheel a power, -1.0 to 1.0. Green pylon.
        """
        motor_values = [float_to_byte(-a),
                        float_to_byte(-b),
                        float_to_byte(-c)]

        self._send(Arduino.DEVICE_MOTORS_SET, motor_values)

    def set_lights(self, hue, saturation, value):
        """
        Set all the neopixel lights to a constant value.

        :param hue:
            0-255 hue
        :param saturation:
            0-255 saturation
        :param value:
            0-255 value
        """
        light_values = [check_byte(hue), check_byte(saturation), check_byte(value)]
        self._send(Arduino.DEVICE_LIGHTS_SET, light_values)

    def get_encoder_values(self):
        """
        Read data from the encoders, returning as a triple of what would be a uint16
        if we had such things.

        :return:
            Triple of encoder values for each wheel.
        """
        encoder_data = self._read(Arduino.DEVICE_ENCODERS_READ, 6)
        return [encoder_data[0] * 256 + encoder_data[1],
                encoder_data[2] * 256 + encoder_data[3],
                encoder_data[4] * 256 + encoder_data[5]]



__author__ = 'tom'


# Classes representing sensors on the robot


class WheelEncoders():
    """
    Class to read absolute values of the quadrature encoders attached to the drive spindle on each wheel motor. These
    encoders are made available over the I2C bus from a microcontroller. This approach is used to allow for more robust
    interrupt driven handling of the encoders using the pin change interrupt vector on the ATMEGA328 chip (in this case
    within an Arduino Nano clone of some description).
    """

    def __init__(self):
        """
        Create a new proxy to the encoders. This doesn't attempt to actually read any data or test connectivity.
        """

    def read(self):
        """
        Read the values of the encoders. Values are stored as two byte unsigned integers in the microcontroller so will
        wrap at 0 and 2^16-1 for the low and high ends. The encoders emit around a thousand events per revolution, so it
        should be reasonably easy to distinguish between a wrap of the value and a genuinely large jump.

        :return:
            A sequence of integers containing the current absolute positions for each wheel
        """



import fcntl
import socket
import struct
from time import time, sleep as time_sleep


class IntervalCheck:
    """
    Utility class which can be used to run code within a polling loop at most once per n seconds. Set up an instance
    of this class with the minimum delay between invocations then enclose the guarded code in a construct such as
    if interval.should_run(): - this will manage the scheduling and ensure that the inner code will only be called if
    at least the specified amount of time has elapsed.

    This class is particularly used to manage hardware where we may wish to include a hardware read or write in a fast
    polling loop such as the task manager, but where the hardware itself cannot usefully be written or read at that
    high rate.

    Instances of this class can also be used in 'with' clauses, i.e. 'with interval:' - this will sleep if required
    before running the gated code, then set the last run time to be the current time. This is not quite the same as
    just calling sleep() before running a code block, as it resets the time after the code has run, instead of after
    the sleep call has completed. Used in this mode therefore the interval is from the end of one code block to the
    start of the next, whereas normally it is from the start of one code block to the start of the next.
    """

    def __init__(self, interval):
        """
        Constructor

        :param float interval:
            The number of seconds that must pass between True values from the should_run() function
        """
        self.interval = interval
        self.last_time = None

    def should_run(self):
        """
        Determines whether the necessary interval has elapsed. If it has, this returns True and updates the internal
        record of the last runtime to be 'now'. If the necessary time has not elapsed this returns False
        """
        now = time()
        if self.last_time is None or now - self.last_time > self.interval:
            self.last_time = now
            return True
        else:
            return False

    def sleep(self):
        """
        Sleep, if necessary, until the minimum interval has elapsed. If the last run time is not set this function will
        set it as a side effect, but will not sleep in this case. Calling sleep() repeatedly will therefore not sleep
        on the first invocation but will subsequently do so each time.
        """
        now = time()
        if self.last_time is None:
            self.last_time = now
            return
        elif now - self.last_time > self.interval:
            return
        else:
            remaining_interval = self.interval - (now - self.last_time)
            time_sleep(remaining_interval)
            self.last_time = now + remaining_interval

    def __enter__(self):
        self.sleep()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.last_time = time()


def in_range(value, min_value, max_value):
    """
    Clamps a value to be within the specified range. If the value is None then None is returned. If either
    max_value or min_value are None they aren't used.

    :param value:
        The value
    :param min_value:
        Minimum allowed value
    :param max_value:
        Maximum allowed value
    """
    if value is None:
        return None
    elif value < min_value and min_value is not None:
        return min_value
    elif value > max_value and max_value is not None:
        return max_value
    return value


def get_ip_address(ifname='wlan0'):
    """
    Get a textual representation of the IP address for the specified interface, defaulting to wlan0 to pick
    up our wireless connection if we have one.

    :param ifname:
        Name of the interface to query, defaults to 'wlan0'

    :return:
        Textual representation of the IP address
    """
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        return socket.inet_ntoa(fcntl.ioctl(
            s.fileno(),
            0x8915,  # SIOCGIFADDR
            struct.pack('256s', ifname[:15])
        )[20:24])
    except:
        return '--.--.--'



import serial
from triangula.util import IntervalCheck, in_range

"""
This is the minimum delay between instructions to the LCD in seconds.
"""
MIN_DELAY = 0.05


class LCD:
    """
    Handle the serial attached LCD display

    To enable this display on the Raspberry Pi you need to first disable the serial console from raspi-config

    """

    def __init__(self):
        """
        Create a new serial connection and clear the display
        """
        self.ser = serial.Serial(port='/dev/ttyAMA0', baudrate=9600)
        self.row1 = ''
        self.row2 = ''
        self.interval = IntervalCheck(interval=MIN_DELAY)
        self.clear()

    def set_text(self, row1=None, row2=None):
        """
        Set the text for either or both rows, send the updated values to the display. In order to fit on our display
        both rows must be at most sixteen characters in length. The display controller actually accepts 40 character
        rows but the display won't display them! This will trim any strings to 16 characters as that's the maximum
        length of our rows

        :param row1:
            Text for row 1
        :param row2:
            Text for row 2
        """
        if row1 is not None:
            self.row1 = row1[:16]
        if row2 is not None:
            self.row2 = row2[:16]
        self._update()

    def set_backlight(self, red=0, green=0, blue=0):
        """
        Set the backlight colour, red green and blue values all range from 0 to 10

        :param red:
            Red value, int 0 to 10
        :param green:
            Green value, int 0 to 10
        :param blue:
            Blue value, int 0 to 10
        """
        with self.interval:
            self._send(
                'pb' +
                str(in_range(red, 0, 10)) + ',' +
                str(in_range(green, 0, 10)) + ',' +
                str(in_range(blue, 0, 10)))

    def clear(self):
        """
        Clear the display
        """
        with self.interval:
            self._send('pc1')

    def cursor_off(self):
        """
        Disable the cursor
        """
        with self.interval:
            self._send('pc12')

    def cursor_blink(self):
        """
        Set the cursor to a flashing block
        """
        with self.interval:
            self._send('pc15')

    def cursor_on(self):
        """
        Set the cursor to a normal underscore character
        """
        with self.interval:
            self._send('pc14')

    def _update(self):
        """
        Update the display with the current text in row1 and row2

        :internal:
        """
        with self.interval:
            self._send('pc2')
            self._send('pd' + self.row1.ljust(40) + self.row2.ljust(16))

    def _send(self, command):
        """
        Send a carriage return terminated message to the display

        :internal:

        :param command:
            string or byte array to send
        """
        self.ser.write(command)
        self.ser.write([0xd])



import sys
import threading
import time

import os

sys.path.append('.')

_imu = None

try:
    import RTIMU

    SETTINGS_FILE = 'RTIMULib'
    print("Using settings file " + SETTINGS_FILE + ".ini")
    if not os.path.exists(SETTINGS_FILE + ".ini"):
        print("Settings file does not exist, will be created")
    s = RTIMU.Settings(SETTINGS_FILE)
    print s
    _imu = RTIMU.RTIMU(s)
    if not _imu.IMUInit():
        raise ImportError('Unable to initialise IMU')
    _pressure = RTIMU.RTPressure(s)
    if not _pressure.pressureInit():
        raise ImportError('Unable to initialise pressure sensor')

    _imu.setSlerpPower(0.02)
    _imu.setGyroEnable(True)
    _imu.setAccelEnable(True)
    _imu.setCompassEnable(True)
    _poll_interval = _imu.IMUGetPollInterval()
    print("Recommended Poll Interval: %dmS\n" % _poll_interval)
    _last_read_time = None
    _last_read_value = None
except ImportError:
    print 'Not importing RTIMU, expected during sphinx generation on OSX'


class IMUThread(threading.Thread):
    def __init__(self):
        super(IMUThread, self).__init__(name='IMU Thread')
        self.fusion_pose = None
        self.setDaemon(daemonic=True)

    def run(self):
        while True:
            if _imu.IMURead():
                data = _imu.getIMUData()
                self.fusion_pose = data["fusionPose"]
                time.sleep(_poll_interval * 2.0 / 1000.0)

if _imu is not None:
    thread = IMUThread()
    thread.start()


def name():
    return 'IMU: {}, Pressure: {}'.format(_imu.IMUName(), _pressure.pressureName())


def read():
    return thread.fusion_pose



__author__ = 'tom'
from setuptools import setup

# Makes use of the sphinx and sphinx-pypi-upload packages. To build for local development
# use 'python setup.py develop'. To upload a version to pypi use 'python setup.py clean sdist upload'.
# To build docs use 'python setup.py build_sphinx' and to upload docs to pythonhosted.org use
# 'python setup.py upload_sphinx'. Both uploads require 'python setup.py register' to be run, and will
# only work for Tom as they need the pypi account credentials.

setup(
    name='triangula',
    version='0.3.1',
    description='Code for Triangula',
    classifiers=['Programming Language :: Python :: 2.7'],
    url='https://github.com/tomoinn/triangula/',
    author='Tom Oinn',
    author_email='tomoinn@gmail.com',
    license='ASL2.0',
    packages=['triangula'],
    install_requires=['evdev==0.5.0', 'euclid==0.1', 'pyserial==2.7', 'numpy==1.10.1'],
    include_package_data=True,
    test_suite='nose.collector',
    tests_require=['nose'],
    dependency_links=[],
    zip_safe=False)
